"""
This awq module has adaption of open-source code from the following sources:
- llm-awq (https://github.com/mit-han-lab/llm-awq?tab=MIT-1-ov-file), licensed under MIT.
"""

from .auto_clip import *
from .auto_scale import *
from .qmodule import *
